<?php
/** @noinspection PhpUnused */
(static function (): void {
	
	$pdfSample = new class('Sample title', 'Me', 'This system', 'fr') {
		protected const USER_SPACE_FACTOR = 595.32001 / 21; // Because 21cm = 595.32001 unit
		private string $docTitle;
		/** @var int[] */
		private array $xrefs;
		private int $objIndex;
		private string $fileData;
		private string $author;
		private string $producer;
		private string $language;
		
		public function __construct(
				string $docTitle,
				string $author,
				string $producer,
				string $language) {
			$this->docTitle = $docTitle;
			$this->author = $author;
			$this->producer = $producer;
			$this->language = $language;
		}
		
		// Utility functions
		final protected function obj(string $content): void {
			$this->xrefs[++$this->objIndex] = strlen($this->fileData);
			$this->fileData .= "{$this->objIndex} 0 obj\r\n" .
					"$content\r\n" .
					"endobj\r\n";
		}
		
		final protected function objEncode(array $entries): string {
			$content = '';
			foreach ($entries as $key => $value) {
				$content .= "/$key $value";
			}
			return "<<$content>>";
		}
		
		final protected function objMap(array $entries, ?string $stream = null): void {
			$this->obj(
					$this->objEncode($entries)
					. ($stream !== null ? "\r\nstream\r\n$stream\r\nendstream" : ""));
		}
		
		/** @noinspection PhpMethodUsage */
		final protected function date(int $time): string {
			return date("YmdHis+00'00", $time);
		}
		
		final protected function stringEscape(string $str): string {
			return str_replace(array("(", ")", "\\"), array("\\(", "\\)", "\\\\"), $str);
		}
		
		final protected function lengthInCm(float $length): float {
			return $length * static::USER_SPACE_FACTOR;
		}
		
		final protected function rectangleInCm(float $width, float $height): string {
			return "[ 0.0 0.0 {$this->lengthInCm($width)} {$this->lengthInCm($height)} ]";
		}
		
		final protected function objReference(int $index): string {
			return "$index 0 R";
		}
		
		final protected function rawString(string $str): string {
			return "({$this->stringEscape($str)})";
		}
		
		// Adds headers, footers and such
		final public function content(): void {
			// Make the PDF file
			$this->xrefs = array();
			$this->objIndex = 0;
			
			$this->fileData = "%PDF-1.7\r\n";
			
			$this->contentBuilder();
			
			// Footer, with Xrefs, trailer and ending line
			$startXref = strlen($this->fileData);
			$xrefCount = $this->objIndex + 1;
			$this->fileData .=
					"xref\r\n" .
					"0 $xrefCount\r\n" .
					"0000000000 65535 f\r\n";
			foreach ($this->xrefs as $xref) {
				$this->fileData .= str_pad($xref, 10, '0', STR_PAD_LEFT) . " 00000 n\r\n";
			}
			
			$this->fileData .= /** @lang Text */
					"trailer\r\n" .
					"<</Info {$this->objIndex} 0 R/Root 1 0 R/Size $xrefCount>>\r\n";
			
			$this->fileData .= /** @lang Text */
					"startxref\r\n" .
					"$startXref\r\n";
			
			$this->fileData .= /** @lang Text */
					"%%EOF\r\n";
			
			echo $this->fileData;
		}
		
		// Content builder, override it
		final protected function contentBuilder(): void {
			// Root node (document catalog)
			$this->objMap(array(
					"Type" => "/Catalog",
					"Pages" => $this->objReference(2),
					"PageLayout" => "/TwoPageLeft",
					"PageMode" => "/UseOutlines",
					"Lang" => "({$this->stringEscape($this->language)})"));
			// Document pages tree (must be a dedicated object refered by above catalog)
			$this->objMap(array(
					"Type" => "/Pages",
					"Count" => 1,
					"Kids" => "[ {$this->objReference(3)} ]"));
			$this->objMap(array(
					"Type" => "/Page",
					"Contents" => "[ {$this->objReference(12)} ]",
					"CropBox" => $this->rectangleInCm(21, 29.7),
					"MediaBox" => $this->rectangleInCm(21, 29.7),
					"Parent" => $this->objReference(2),
					"Resources" => $this->objReference(13),
					"Rotate" => "0"));
			$this->obj($this->rawString("Identity"));
			$this->obj($this->rawString("Adobe"));
			$this->obj("[ -549 -250 549 919 ]");
			$this->obj("549");
			
			//		$fontsource = file_get_contents(__DIR__. '/a-space-regular-demo.otf');
			$fontsource = file_get_contents(__DIR__ . '/obj8.ttf');
			$fontdeflate = gzdeflate($fontsource, 9, ZLIB_ENCODING_DEFLATE);
			$this->objMap(
					array(
							"Type" => "/Stream",
							"Filter" => "/FlateDecode",
							"Length" => strlen($fontdeflate),
							"Length1" => strlen($fontsource)),
					$fontdeflate);
			$this->obj("[ 3 3 549 11 11 549 26 26 549 134 134 549 135 135 549 142 142 549 145 145 549 148 148 549 ]");
			$this->objMap(
					array("Length" => 436),
					/** @lang Text */
					"/CIDInit /ProcSet findresource begin 12 dict begin begincmap /CIDSystemInfo << /Registry (Adobe) /Ordering (UCS) /Supplement 0 >> def /CMapName /Adobe-Identity-UCS def /CMapType 2 def 1 begincodespacerange <0000> <FFFF> endcodespacerange 8 beginbfchar <0003> <0020> <000B> <0048> <001A> <0057> <0086> <0064> <0087> <0065> <008E> <006C> <0091> <006F> <0094> <0072> endbfchar endcmap CMapName currentdict /CMap defineresource pop end end ");
			
			$this->objMap(array(
					"BaseFont" => "/CIDFont+F1",
					"DescendantFonts" =>
							"[ "
							. $this->objEncode(array(
									"BaseFont" => "/CIDFont+F1",
									"CIDSystemInfo" => $this->objEncode(array(
											"Ordering" => $this->objReference(4),
											"Registry" => $this->objReference(5),
											"Supplement" => 0)),
									"CIDToGIDMap" => "/Identity",
									"FontDescriptor" => $this->objEncode(array(
											"Ascent" => 919,
											"CapHeight" => 638,
											"Descent" => -250,
											"Flags" => 6,
											"FontBBox" => $this->objReference(6),
											"FontFile2" => $this->objReference(8),
											"FontName" => "/CIDFont+F1",
											"ItalicAngle" => 0,
											"StemV" => $this->objReference(7),
											"Type" => "/FontDescriptor")),
									"Subtype" => "/CIDFontType2",
									"Type" => "/Font",
									"W" => $this->objReference(9)))
							. " ]",
					"Encoding" => "/Identity-H",
					"Subtype" => "/Type0",
					"ToUnicode" => $this->objReference(10),
					"Type" => "/Font"));
			
			$pageContent =
					"0.75 0 0 -0.75 0 {$this->lengthInCm(29.7)} cm\r\n" .
					"q\r\n" .
					"0.000000 0.000000 0.000000 rg\r\n" .
					"BT\r\n" .
					"0 Tr\r\n" .
					"/F1 13 Tf\r\n" .
					"1 0 0 -1 {$this->lengthInCm(1)} {$this->lengthInCm(1)} Tm\r\n" .
					"[" .
					"<000B>" .
					"<0087>" .
					"<008E>" .
					"<008E>" .
					"<0091>" .
					"<0003>" .
					"<001A>" .
					"<0091>" .
					"<0094>" .
					"<008E>" .
					"<0086>" .
					"] TJ\r\n" .
					"ET\r\n" .
					"Q\r\n";
			$this->objMap(
					array(
							"Filter" => "/FlateDecode",
							"Length" => strlen($pageContent)),
					gzdeflate($pageContent, 5, ZLIB_ENCODING_DEFLATE));
			$this->objMap(array("Font" => $this->objEncode(array("F1" => $this->objReference(11)))));
			
			// Document's metadata
			$this->objMap(array(
					"Author" => $this->rawString($this->author),
					"CreationDate" => $this->rawString(time()),
					"ModDate" => $this->rawString(time()),
					"Producer" => $this->rawString($this->producer),
					"Title" => $this->rawString($this->docTitle)));
		}
	};
	
	header('Content-Type: application/pdf');
	ob_start();
	$pdfSample->content();
	$content = ob_get_clean();
	file_put_contents(__DIR__ . '/document.pdf', $content);
	echo $content;
})();
